"""Typing module."""
from __future__ import annotations

from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

PyrgoProgram: TypeAlias = Literal[
    "ruff",
    "mypy.dmypy",
    "piptools",
    "pip",
    "build",
    "pytest",
    "pip_audit",
    "vulture",
    "pdoc",
]



"""Lock command."""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Literal

import click
from result import Ok, Result
from typing_extensions import assert_never

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program

if TYPE_CHECKING:
    import subprocess


def _initial_args(
    env: str,
    core_deps_alias: str,
    *,
    upgrade: bool,
) -> list[str]:
    base_args = ["compile"]
    if upgrade:
        base_args.append("-U")

    if env == core_deps_alias:
        return base_args

    base_args.extend(["--no-strip-extras", "--extra", env])
    return base_args


def _complete_cmd(
    *,
    env: str,
    cmd: PythonCommandExec,
    generate_hashes: bool,
    config: PyrgoConf,
) -> PythonCommandExec:
    if generate_hashes:
        cmd.add_args(args=["--generate-hashes"])

    cmd.add_args(
        args=[
            "--strip-extras",
            "--resolver=backtracking",
            "-o",
            config.requirements.joinpath(f"{env}.txt")
            .relative_to(config.cwd)
            .as_posix(),
            "pyproject.toml",
        ],
    )
    return cmd


@click.command("lock")
@click.option(
    "--generate-hashes",
    "generate_hashes",
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
)
@click.option(
    "--upgrade",
    "upgrade",
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
)
@click.option(
    "-e",
    "--env",
    "envs",
    multiple=True,
    type=click.STRING,
    required=False,
)
def lock(*, generate_hashes: bool, envs: tuple[str, ...], upgrade: bool) -> None:
    """Lock project dependencies with `piptools`."""
    configuration = PyrgoConf()

    if not configuration.requirements.exists():
        configuration.requirements.mkdir()

    execution_mode: Literal["all", "specific-groups"] = (
        "all" if len(envs) == 0 else "specific-groups"
    )
    program_execution: Result[None, list[subprocess.CalledProcessError]]
    all_commands: list[PythonCommandExec] = []
    if execution_mode == "all":
        all_commands.extend(
            _complete_cmd(
                cmd=PythonCommandExec(
                    program="piptools",
                ).add_args(
                    args=_initial_args(
                        env=env,
                        core_deps_alias=configuration.env_groups[0],
                        upgrade=upgrade,
                    ),
                ),
                generate_hashes=generate_hashes,
                config=configuration,
                env=env,
            )
            for env in configuration.env_groups
        )

        program_execution = inform_and_run_program(commands=all_commands)

    elif execution_mode == "specific-groups":
        for env in envs:
            if env in configuration.env_groups:
                all_commands.append(
                    _complete_cmd(
                        cmd=PythonCommandExec(
                            program="piptools",
                        ).add_args(
                            args=_initial_args(
                                env=env,
                                core_deps_alias=configuration.env_groups[0],
                                upgrade=upgrade,
                            ),
                        ),
                        generate_hashes=generate_hashes,
                        config=configuration,
                        env=env,
                    ),
                )

            else:
                click.echo(
                    message=click.style(
                        text=f"Optional deps for `{env}` not found in pyproject.toml.",
                        fg="red",
                    ),
                )
                sys.exit(1)

        program_execution = inform_and_run_program(commands=all_commands)
    else:
        assert_never(execution_mode)

    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Fix command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program


@click.command("fix")
def fix() -> None:
    """Automatically fix lint warnings reported by `ruff`."""
    configuration = PyrgoConf()
    ruff_command = (
        PythonCommandExec(
            program="ruff",
        )
        .add_args(args=["--unsafe-fixes", "--fix"])
        .add_args(configuration.relevant_paths)
    )

    program_execution = inform_and_run_program(
        commands=[
            ruff_command,
        ]
    )
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Check command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program


def _build_vulture_cmd(
    *,
    add_noqa: bool,
    ignore_noqa: bool,
    configuration: PyrgoConf,
    vulture_allowlist: str,
) -> PythonCommandExec:
    configuration.cwd.joinpath(vulture_allowlist).touch(exist_ok=True)
    vulture_command = PythonCommandExec(
        program="vulture",
    )
    if add_noqa:
        vulture_command.add_args(args=["--make-whitelist"]).add_output_file(
            file=configuration.cwd.joinpath(vulture_allowlist)
        )

    vulture_command.add_args(configuration.relevant_paths)
    if not ignore_noqa:
        vulture_command.add_args(args=[vulture_allowlist])

    return vulture_command


@click.command("check")
@click.option(
    "-t",
    "--timeout",
    "timeout",
    default=360,
    type=click.IntRange(min=60),
)
@click.option(
    "--add-noqa",
    "add_noqa",
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
)
@click.option(
    "--ignore-noqa",
    "ignore_noqa",
    is_flag=True,
    default=False,
    show_default=True,
    type=click.BOOL,
)
def check(*, timeout: int, add_noqa: bool, ignore_noqa: bool) -> None:
    """Check code with `mypy`, `ruff` and `vulture`."""
    configuration = PyrgoConf()
    ruff_command = PythonCommandExec(
        program="ruff",
    )
    mypy_command = PythonCommandExec(
        program="mypy.dmypy",
    ).add_args(
        args=["run", "--timeout", str(timeout), "--"],
    )

    if add_noqa:
        ruff_command.add_args(args=["--add-noqa"])
    if ignore_noqa:
        ruff_command.add_args(args=["--ignore-noqa"])

    ruff_command.add_args(configuration.relevant_paths)
    mypy_command.add_args(configuration.relevant_paths)

    program_execution = inform_and_run_program(
        commands=[
            ruff_command,
            mypy_command,
            _build_vulture_cmd(
                add_noqa=add_noqa,
                ignore_noqa=ignore_noqa,
                configuration=configuration,
                vulture_allowlist=configuration.vulture_allowlist.as_posix(),
            ),
        ]
    )
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""CLI entrypoint."""
from __future__ import annotations

import click

from pyrgo.cmds.add import add
from pyrgo.cmds.audit import audit
from pyrgo.cmds.build import build
from pyrgo.cmds.check import check
from pyrgo.cmds.clean import clean
from pyrgo.cmds.doc import doc
from pyrgo.cmds.fix import fix
from pyrgo.cmds.fmt import fmt
from pyrgo.cmds.lock import lock
from pyrgo.cmds.new import new
from pyrgo.cmds.remove import remove
from pyrgo.cmds.sync import sync
from pyrgo.cmds.test import test


@click.group(
    context_settings={
        "help_option_names": [
            "-h",
            "--help",
        ],
        "show_default": True,
    },
)
@click.version_option(None, "-v", "--version")
def cli() -> None:
    """pyrgo. Python package manager."""


cli.add_command(cmd=lock)
cli.add_command(cmd=fmt)
cli.add_command(cmd=check)
cli.add_command(cmd=sync)
cli.add_command(cmd=build)
cli.add_command(cmd=clean)
cli.add_command(cmd=test)
cli.add_command(cmd=audit)
cli.add_command(cmd=new)
cli.add_command(cmd=doc)
cli.add_command(cmd=fix)
cli.add_command(cmd=add)
cli.add_command(cmd=remove)



"""Test command."""
from __future__ import annotations

import sys
from typing import Optional

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.utils import inform_and_run_program


@click.command("test")
@click.option(
    "-m",
    "marker",
    type=click.STRING,
    default=None,
    show_default=True,
)
def test(marker: Optional[str]) -> None:
    """Run tests with `pytest`."""
    pytest_command = PythonCommandExec(
        program="pytest",
    )
    if marker is not None:
        pytest_command.add_args(args=["-m", marker])
    program_execution = inform_and_run_program(commands=[pytest_command])
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Remove command."""
from __future__ import annotations

import click
import tomlkit
from packaging.requirements import Requirement
from tomlkit.items import Item, Table

from pyrgo.conf import PyrgoConf
from pyrgo.utils import ensure_env_exist


@click.command("remove")
@click.option(
    "-e",
    "--env",
    "env",
    type=click.STRING,
    required=True,
    default="core",
    show_default=True,
)
@click.argument("dependency", type=click.STRING)
def remove(env: str, dependency: str) -> None:
    """Remove dependency from env."""
    conf = PyrgoConf()
    ensure_env_exist(env=env, config=conf, where="pyproject")
    pyproject_doc = tomlkit.loads(conf.pyproject_path.read_text())

    project_container = pyproject_doc["project"]
    if not isinstance(project_container, Table):
        msg = "`project` section should be a container"
        raise TypeError(msg)

    if env == "core":
        project_container["dependencies"] = _remove_from_pyproject_list(
            project_container["dependencies"], value_to_remove=dependency
        )

    else:
        project_optional_deps = project_container["optional-dependencies"]
        if not isinstance(project_optional_deps, Table):
            msg = "`project` section should be a container"
            raise TypeError(msg)

        new_deps = _remove_from_pyproject_list(
            pyproject_item=project_optional_deps[env], value_to_remove=dependency
        )
        if len(new_deps) > 0:
            project_optional_deps[env] = new_deps
        else:
            project_optional_deps.pop(env)

    conf.pyproject_path.write_text(tomlkit.dumps(pyproject_doc, sort_keys=False))


def _remove_from_pyproject_list(
    pyproject_item: Item, value_to_remove: str
) -> list[str]:
    pyproject_dependencies: dict[str, str] = {
        Requirement(x).name: x for x in pyproject_item.unwrap()
    }
    if value_to_remove in pyproject_dependencies:
        pyproject_dependencies.pop(value_to_remove)

    return list(pyproject_dependencies.values())



"""Format command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program


@click.command("fmt")
def fmt() -> None:
    """Format code with `ruff`."""
    configuration = PyrgoConf()
    fmt_command = PythonCommandExec(
        program="ruff",
    ).add_args(args=["format", "."])
    for command in [fmt_command]:
        command.add_args(args=configuration.relevant_paths)

    program_execution = inform_and_run_program(commands=[fmt_command])
    if not isinstance(program_execution, Ok):
        sys.exit(1)
    sys.exit(0)



"""Add command."""
from __future__ import annotations

import click
import tomlkit
from tomlkit.items import Item, Table

from pyrgo.conf import PyrgoConf


@click.command("add")
@click.option(
    "-e",
    "--env",
    "env",
    type=click.STRING,
    required=True,
    default="core",
    show_default=True,
)
@click.argument("dependency", type=click.STRING)
def add(env: str, dependency: str) -> None:
    """Add dependency to env."""
    conf = PyrgoConf()

    pyproject_doc = tomlkit.loads(conf.pyproject_path.read_text())

    project_container = pyproject_doc["project"]
    if not isinstance(project_container, Table):
        msg = "`project` section should be a container"
        raise TypeError(msg)

    if env == "core":
        project_container["dependencies"] = _extend_pyproject_list(
            project_container["dependencies"], new_value=dependency
        )

    else:
        project_optional_deps = project_container["optional-dependencies"]
        if not isinstance(project_optional_deps, Table):
            msg = "`project` section should be a container"
            raise TypeError(msg)

        if env not in conf.env_groups:
            project_optional_deps[env] = []

        project_optional_deps[env] = _extend_pyproject_list(
            pyproject_item=project_optional_deps[env], new_value=dependency
        )

    conf.pyproject_path.write_text(tomlkit.dumps(pyproject_doc, sort_keys=False))


def _extend_pyproject_list(pyproject_item: Item, new_value: str) -> list[str]:
    pyproject_list: list[str] = pyproject_item.unwrap()
    pyproject_list.append(new_value)
    return pyproject_list



"""New command."""
from __future__ import annotations

import pathlib
import shutil
import sys

import click

from pyrgo import resources


@click.command("new")
@click.option("-n", "--name", "name", type=click.STRING, required=True)
@click.option("--path", "path", type=click.Path(), required=True)
def new(name: str, path: str) -> None:
    """Create a project."""
    path_path = pathlib.Path(path)

    package_name = name.strip().replace("-", "_")
    dir_components = [
        path_path.joinpath(package_name),
        path_path.joinpath("tests"),
    ]
    file_componens = [
        path_path.joinpath("README.md"),
        path_path.joinpath("pyproject.toml"),
    ]

    existing_dirs: list[str] = [
        dir_comp.as_posix()
        for dir_comp in dir_components
        if dir_comp.exists() and dir_comp.is_dir()
    ]
    existing_files: list[str] = [
        file_comp.as_posix()
        for file_comp in file_componens
        if file_comp.exists() and file_comp.is_file()
    ]

    if len(existing_dirs) > 0 or len(existing_files) > 0:
        if len(existing_dirs) > 0:
            click.echo(
                click.style(
                    f"Directories: {existing_dirs} already exists in path {path}",
                    fg="red",
                ),
            )
        if len(existing_files) > 0:
            click.echo(
                click.style(
                    f"Files: {existing_files} already exists in path {path}",
                    fg="red",
                ),
            )
        sys.exit(1)

    shutil.copytree(
        src=pathlib.Path(resources.__file__).parent.joinpath("new-project"),
        dst=path_path,
        dirs_exist_ok=True,
    )

    shutil.move(
        src=path_path.joinpath("new_project"),
        dst=path_path.joinpath(package_name),
    )
    file_paths = [
        path_path.joinpath("README.md"),
        path_path.joinpath("pyproject.toml"),
    ]
    for file_path in file_paths:
        new_data = file_path.read_text().replace("new-project", name)
        file_path.write_text(new_data)



"""Pyrgo configuration."""
from __future__ import annotations

import pathlib
import sys
from dataclasses import dataclass, field
from typing import Any

import click
from mashumaro import field_options
from mashumaro.mixins.toml import DataClassTOMLMixin


@dataclass(frozen=True)
class _Project:
    name: str
    optional_dependencies: dict[str, Any] = field(
        metadata=field_options(alias="optional-dependencies"), default_factory=dict
    )


@dataclass(frozen=True)
class _Pyrgo:
    extra_paths: list[str] = field(
        metadata=field_options(alias="extra-paths"), default_factory=list
    )
    extra_caches: list[str] = field(
        metadata=field_options(alias="extra-caches"), default_factory=list
    )
    vulture_allowlist: str = field(
        metadata=field_options(alias="vulture-allowlist"), default=".whitelist.vulture"
    )


@dataclass(frozen=True)
class _Tooling:
    pytest: dict[str, Any]
    pyrgo: _Pyrgo = field(default=_Pyrgo())


@dataclass(frozen=True)
class _PyProjectToml(DataClassTOMLMixin):
    project: _Project
    tool: _Tooling


class PyProjectNotFoundError(Exception):
    """Raised when `pyproject.toml` not found."""

    def __init__(self, path: pathlib.Path) -> None:  # noqa: D107
        super().__init__(f"`pyproject.toml` not found at {path.as_posix()}")


class PyrgoConf:
    """Pyrgo configuration."""

    def __init__(self) -> None:
        """Create new configuration instance."""
        cwd = pathlib.Path().cwd()
        pyproject_path = cwd.joinpath("pyproject.toml")
        if not (pyproject_path.exists() and pyproject_path.is_file()):
            raise PyProjectNotFoundError(path=cwd)

        pyproject_data = _PyProjectToml.from_toml(pyproject_path.read_text())

        project_name = pyproject_data.project.name.strip().replace("-", "_")
        relevant_paths: list[str] = [
            project_name,
        ]
        try:
            test_paths = pyproject_data.tool.pytest["ini_options"]["testpaths"]

        except KeyError:
            click.echo(
                message=click.style(
                    "`tool.pytest.ini_options.testpaths` is required.", fg="red"
                ),
                color=True,
            )
            sys.exit(1)

        caches = [
            cwd.joinpath(".pytest_cache"),
            cwd.joinpath(".ruff_cache"),
            cwd.joinpath(".mypy_cache"),
            *(cwd.joinpath(extra) for extra in pyproject_data.tool.pyrgo.extra_caches),
        ]

        relevant_paths.extend(
            test_paths,
        )
        relevant_paths.extend(pyproject_data.tool.pyrgo.extra_paths)

        self.cwd: pathlib.Path = cwd
        self.requirements: pathlib.Path = cwd.joinpath("requirements")
        self.relevant_paths: list[str] = relevant_paths
        self.artifacts: list[pathlib.Path] = [cwd.joinpath("dist")]
        self.caches: list[pathlib.Path] = caches
        self.env_groups: list[str] = [
            "core",
            *pyproject_data.project.optional_dependencies.keys(),
        ]
        self.project_name: str = project_name
        self.pyproject_path: pathlib.Path = pyproject_path
        self.vulture_allowlist: pathlib.Path = cwd.joinpath(
            pyproject_data.tool.pyrgo.vulture_allowlist
        ).relative_to(cwd)

    def locked_envs(self) -> set[str]:
        """Get a set of available locked envs."""
        locked_envs: set[str] = set()
        for env in self.requirements.glob(pattern="*.txt"):
            locked_envs.add(
                env.relative_to(self.requirements).as_posix().removesuffix(".txt"),
            )

        return locked_envs



"""Doc command."""
from __future__ import annotations

from typing import Optional

import click

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program


@click.command("doc")
@click.option(
    "-o",
    "output_dir",
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
    ),
)
@click.option(
    "-p",
    "port",
    type=click.INT,
)
def doc(output_dir: Optional[str], port: Optional[int]) -> None:
    """Build a package's documentation with `pdoc`."""
    configuration = PyrgoConf()
    pdoc_command = PythonCommandExec(program="pdoc").add_args(
        args=[
            configuration.project_name,
        ]
    )
    if port is not None:
        pdoc_command.add_args(args=["-p", str(port)])

    if output_dir is not None:
        pdoc_command.add_args(args=["-o", output_dir])

    inform_and_run_program(commands=[pdoc_command])



"""CLI utilities."""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Literal

import click
from result import Err, Ok, Result
from typing_extensions import assert_never

if TYPE_CHECKING:
    import subprocess

    from pyrgo.command_exec import PythonCommandExec
    from pyrgo.conf import PyrgoConf


def inform_and_run_program(
    commands: list[PythonCommandExec],
) -> Result[None, list[subprocess.CalledProcessError]]:
    """Inform and execute command."""
    subprocess_errors: list[subprocess.CalledProcessError] = []
    for command in commands:
        command_to_execute = " ".join(command.args)
        click.echo(message=click.style(command_to_execute, fg="yellow"), color=True)
        execution_result = command.execute()
        if not isinstance(execution_result, Ok):
            subprocess_errors.append(execution_result.err())

    if len(subprocess_errors) > 0:
        return Err(subprocess_errors)
    return Ok(None)


def ensure_env_exist(
    env: str, config: PyrgoConf, where: Literal["lock-files", "pyproject"]
) -> None:
    """Ensure selected env exists in `requirements` folder."""
    if where == "lock-files":
        if not config.requirements.exists():
            click.echo(
                message=click.style(
                    "No `requirements` folder found. Try locking your deps first",
                    fg="red",
                ),
            )
            sys.exit(1)

        locked_envs = config.locked_envs()
        if env not in locked_envs:
            click.echo(
                click.style(
                    f"`{env}` not found in available envs.\nAvailable envs: {locked_envs}",
                    fg="red",
                ),
            )

            sys.exit(1)
    elif where == "pyproject":
        if env not in config.env_groups:
            click.echo(
                message=click.style(
                    f"Environment `{env}` not present in `pyproject.toml`",
                    fg="red",
                )
            )
    else:
        assert_never(where)



"""Sync command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import ensure_env_exist, inform_and_run_program


@click.command("sync")
@click.option(
    "-e",
    "--env",
    "env",
    type=click.STRING,
    required=True,
)
@click.option(
    "--editable/--no-editable",
    "editable",
    type=click.BOOL,
    default=True,
    show_default=True,
)
def sync(env: str, *, editable: bool) -> None:
    """Sync current python environment to locked deps."""
    config = PyrgoConf()
    ensure_env_exist(env=env, config=config, where="lock-files")

    piptools_command = PythonCommandExec(
        program="piptools",
    ).add_args(
        args=[
            "sync",
            config.requirements.joinpath(f"{env}.txt")
            .relative_to(config.cwd)
            .as_posix(),
        ],
    )
    pip_command = PythonCommandExec(
        program="pip",
    ).add_args(
        args=["install", "--no-deps"],
    )
    if editable:
        pip_command.add_args(args=["-e"])
    pip_command.add_args(args=["."])

    program_execution = inform_and_run_program(
        commands=[
            piptools_command,
            pip_command,
        ]
    )

    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Command execution module."""
from __future__ import annotations

import subprocess
import sys
from typing import TYPE_CHECKING, Optional

from result import Err, Ok, Result

if TYPE_CHECKING:
    import pathlib
    from io import TextIOWrapper
    from pathlib import Path

    from typing_extensions import Self

    from pyrgo.typing import PyrgoProgram


class PythonCommandExec:
    """Python command executor."""

    def __init__(
        self,
        program: PyrgoProgram,
    ) -> None:
        """Build a new command executor."""
        self.args: list[str] = [sys.executable, "-m", program]
        self.output_file: Path | None = None

    def add_output_file(self, file: pathlib.Path) -> Self:
        """Add output file to command."""
        self.output_file = file
        return self

    def add_args(self, args: list[str]) -> Self:
        """Add arguments to command."""
        self.args.extend(args)
        return self

    def execute(self) -> Result[None, subprocess.CalledProcessError]:
        """Execute python command."""
        stdout_file: Optional[TextIOWrapper] = None

        if self.output_file is not None:
            stdout_file = self.output_file.open(mode="w")

        try:
            subprocess.run(args=self.args, check=True, stdout=stdout_file)
        except subprocess.CalledProcessError as e:
            return Err(e)
        return Ok(None)



"""Build command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program


@click.command("build")
def build() -> None:
    """Build project with `build`."""
    program_execution = inform_and_run_program(
        commands=[
            PythonCommandExec(
                program="build",
            ).add_args(
                [PyrgoConf().cwd.as_posix()],
            ),
        ],
    )
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)



"""Clean command."""
from __future__ import annotations

import itertools
import shutil
import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import inform_and_run_program


@click.command("clean")
def clean() -> None:
    """Clean project repository."""
    config = PyrgoConf()
    for folder in itertools.chain(config.artifacts, config.caches):
        if folder.exists():
            if folder.is_dir():
                shutil.rmtree(path=folder)
            else:
                folder.unlink()

    for path in config.relevant_paths:
        for pycache in config.cwd.joinpath(path).rglob(pattern="__pycache__"):
            shutil.rmtree(pycache)

    for outdated_file in config.locked_envs().difference(config.env_groups):
        config.requirements.joinpath(f"{outdated_file}.txt").unlink()

    program_execution = inform_and_run_program(
        commands=[
            PythonCommandExec(
                program="mypy.dmypy",
            ).add_args(args=["stop"])
        ],
    )
    if not isinstance(program_execution, Ok):
        sys.exit(1)
    sys.exit(0)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import pytest

from pyrgo.command_exec import PythonCommandExec

if TYPE_CHECKING:
    from pyrgo.typing import PyrgoProgram


@pytest.mark.unit()
class TestPythonCommandExec:
    @pytest.mark.parametrize(argnames="program", argvalues=["ruff", "pip"])
    def test_new(self, program: PyrgoProgram) -> None:
        assert PythonCommandExec(program=program).args == [
            sys.executable,
            "-m",
            program,
        ]

    @pytest.mark.parametrize(
        argnames=["command", "args", "expected"],
        argvalues=[
            (
                PythonCommandExec(program="pip"),
                ["1", "2"],
                ["pip", "1", "2"],
            ),
        ],
    )
    def test_add_args(
        self,
        command: PythonCommandExec,
        args: list[str],
        expected: list[str],
    ) -> None:
        base_args = [sys.executable, "-m"]
        base_args.extend(expected)
        assert command.add_args(args=args).args == base_args



"""Audit command."""
from __future__ import annotations

import sys

import click
from result import Ok

from pyrgo.command_exec import PythonCommandExec
from pyrgo.conf import PyrgoConf
from pyrgo.utils import ensure_env_exist, inform_and_run_program


@click.command("audit")
@click.option(
    "-e",
    "--env",
    "env",
    type=click.STRING,
    required=True,
)
@click.option(
    "--fix/--no-fix",
    "fix",
    type=click.BOOL,
    default=False,
    show_default=True,
)
def audit(env: str, *, fix: bool) -> None:
    """Audit locked dependencies with `pip_audit`."""
    config = PyrgoConf()
    ensure_env_exist(env=env, config=config, where="lock-files")

    pip_audit_cmd = PythonCommandExec(
        program="pip_audit",
    ).add_args(
        args=[
            "-r",
            config.requirements.joinpath(f"{env}.txt")
            .relative_to(config.cwd)
            .as_posix(),
        ],
    )
    if fix:
        pip_audit_cmd.add_args(args=["--fix"])

    program_execution = inform_and_run_program(commands=[pip_audit_cmd])
    if not isinstance(program_execution, Ok):
        sys.exit(1)

    sys.exit(0)
